/*
 *  TriMeshOpenGLViewer-Phong
 *  Copyright (C) 2005  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <GL/glu.h>

#include "Camera.h"
#include "Vertex.h"
#include "Triangle.h"

#define XX 0
#define YY 1
#define ZZ 2

Camera::Camera( 
    coord eyex, coord eyey, coord eyez,
    coord lAng, coord uAng, 
    Scene* scene, 
    Screen3D *screen)
    :
    Point(eyex, eyez, eyez),
    _leftangle(lAng), _upangle(uAng),
    _scene(scene),
    _screen(screen),
    _state(RENDER_LINES)
{
    _scene->addCamera(this);
}

void Camera::lookGLU()
{
#if 1
    gluLookAt(
        getX(), getY(), getZ(),
        getX()-1.*cos(_upangle*3.14159/180.0)*cos(_leftangle*3.14159/180.0),
        getY()-1.*cos(_upangle*3.14159/180.0)*sin(_leftangle*3.14159/180.0),
        getZ()+1.*sin(_upangle*3.14159/180.0),
        0.,0.,1.);
#else
    // Disabled... I'm still missing something :-)
    glRotatef(-_upangle, 1.,0.,0.);
    glRotatef(-_leftangle, 0.,1.,0.);
    glTranslatef(-getZ(), -getY(), -getX());
    glRotatef(-90., 1., 0., 0.);
    glRotatef(-90., 0., 0., 1.);
#endif

}

void Camera::renderPoints()
{
    _scene->renderPoints(*this);
}

void Camera::renderLines()
{
    _scene->renderGouraud(*this, true);
}

void Camera::renderGouraud()
{
    _scene->renderGouraud(*this);
}

void Camera::renderPhong()
{
    if (_scene->_lights.size() != 1) {
	cerr << "Phong fragment shader only works with one light!\n";
	cerr << "Switching to Gouraud shading..." << endl;
	setRenderState(RENDER_GOURAUD);
	return;
    }
    _scene->renderPhong(*this);
}

void Camera::render()
{
    switch(_state) {
    case RENDER_POINTS:
        renderPoints();
        break;
    case RENDER_LINES:
        renderLines();
        break;
    case RENDER_GOURAUD:
        renderGouraud();
        break;
    case RENDER_PHONG:
        renderPhong();
        break;
    default:
        THROW("Camera: state not set");
    }
}

void Camera::setRenderState(RenderState state)
{
    _state = state;
    glEnable(GL_NORMALIZE);
    switch(_state) {
    case RENDER_POINTS:
        glDisable(GL_TEXTURE_GEN_S);
        glDisable(GL_TEXTURE_GEN_T);
        glDisable(GL_TEXTURE_GEN_R);
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_LIGHTING);
        break;
    case RENDER_LINES:
        glEnable(GL_CULL_FACE);
        glEnable(GL_LIGHTING);
        glEnable(GL_POLYGON_SMOOTH);
        glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
        break;
    case RENDER_GOURAUD:
        glEnable(GL_CULL_FACE);
        glEnable(GL_LIGHTING);
        glEnable(GL_POLYGON_SMOOTH);
        glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
        break;
    case RENDER_PHONG:
        glEnable(GL_CULL_FACE);
        glEnable(GL_LIGHTING);
        glDisable(GL_POLYGON_SMOOTH);
        glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
        break;
    }

    // No global Ambient (i.e. only ambient from light sources' ambient)
    GLfloat globalAmbient[] = {0.0, 0.0, 0.0, 1.0f};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globalAmbient);

    // Common specular: full metal!
    GLfloat specular[] = { 1.0, 1.0, 1.0, 1.0 };
    glMaterialfv( GL_FRONT, GL_SPECULAR, specular);

    // Midway... max is 128
    GLfloat shine = 64;
    glMaterialfv( GL_FRONT, GL_SHININESS, &shine );
}
