/*
 *  TriMeshOpenGLViewer-Phong
 *  Copyright (C) 2005  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef _WIN32
#include <windows.h>
#else
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glxext.h>
#endif

#include <vector>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <cstdlib>

#include <assert.h>

#include "Object.h"
#include "Vertex.h"
#include "Triangle.h"
#include "Exceptions.h"
#include "Defines.h"
#include "Variables.h"
#include "Normal.h"

Object::Object(	const char *filename, coord x, coord y, coord z)
    :
    _x(x), _y(y), _z(z), 
    _rx(0), _ry(0), _rz(0), 
    _scene(NULL)
{
    Uint32 magic = 0;
    _maxx = _maxy = _maxz = -1.e10;
    _minx = _miny = _minz = 1.e10;

    try {
	const char *dot = strrchr(filename, '.');
	if (dot)
	    dot++;

	if (dot && (!strcmp(dot, "PLY") || !strcmp(dot, "ply"))) {
	    ifstream file(filename, ios::in);
	    if (!file) {
		THROW((string("Missing ")+string(filename)).c_str());
	    }
	    string line;
	    unsigned totalVertices, totalTriangles, lineNo=0;
	    bool inside = false;
	    while(getline(file, line)) {
		lineNo++;
		if (!inside) {
		    if (line.substr(0, 14) == "element vertex") {
			istringstream str(line);
			string word1;
			str >> word1;
			str >> word1;
			str >> totalVertices;
			_vertexes.reserve(totalVertices);
		    } else if (line.substr(0, 12) == "element face") {
			istringstream str(line);
			string word1;
			str >> word1;
			str >> word1;
			str >> totalTriangles;
			_vertexes.reserve(totalTriangles*3);
		    } else if (line.substr(0, 10) == "end_header") 
			inside = true;
		} else {
		    if (totalVertices) {
			totalVertices--;
			coord x,y,z;
			float nx=0.,ny=0.,nz=0.;
			unsigned ambientOcclusionCoeff,dummy;
			istringstream str(line);
			str >> x >> y >> z >> ambientOcclusionCoeff;
			_vertexes.push_back(new Vertex(x,y,z, this, 0.,0.,0., ambientOcclusionCoeff/255.0));
		    } else if (totalTriangles) {
			totalTriangles--;
			unsigned dummy, r, g, b;
			unsigned idx1, idx2, idx3;
			istringstream str(line);
			if (str >> dummy >> idx1 >> idx2 >> idx3)
			{
			    if (str >> r >> g >> b)
				;
			    else {
				r = 255;
				g = 255; 
				b = 255;
			    }
			    _triangles.push_back(
				new Triangle(_vertexes[idx1], _vertexes[idx2], _vertexes[idx3], r/255.0, g/255.0, b/255.0));
			}
		    }
		}
	    }
	    fixNormals();
	} else if (dot && (!strcmp(dot, "tri") || !strcmp(dot, "TRI"))) {
	    FILE *fp = fopen(filename, "rb");
	    if (!fp) {
		THROW("File could not be opened");
	    }

	    fread(&magic, 1, sizeof(Uint32), fp);
	    if (magic != TRI_MAGIC && magic != TRI_MAGICNORMAL) {
		fseek(fp, 0, SEEK_SET);
	    }

	    Uint32 totalPoints = 0, totalTris = 0;
	    do {
		coord x,y,z;
		Uint32 idx1,idx2,idx3;
		Uint32 noOfPoints, noOfTris;

		fread(&noOfPoints, 1, sizeof(Uint32), fp);
		if (feof(fp))
		    break;
		for(Uint32 i=0; i<noOfPoints; i++) {
		    fread(&x,1,4,fp);
		    if (feof(fp)) { THROW("Malformed 3D file"); }
		    fread(&y,1,4,fp);
		    if (feof(fp)) { THROW("Malformed 3D file"); }
		    fread(&z,1,4,fp);
		    if (feof(fp)) { THROW("Malformed 3D file"); }
		    coord nx,ny,nz;
		    if (magic == TRI_MAGICNORMAL) {
			fread(&nx,1,4,fp);
			if (feof(fp)) { THROW("Malformed 3D file (r)"); }
			fread(&ny,1,4,fp);
			if (feof(fp)) { THROW("Malformed 3D file (g)"); }
			fread(&nz,1,4,fp);
			if (feof(fp)) { THROW("Malformed 3D file (b)"); }
		    } else {
			nx = ny = nz = 0;
		    }
		    Vertex *tmp = new Vertex(x,y,z,this,nx,ny,nz);
		    _vertexes.push_back(tmp);
		}

		fread(&noOfTris, 1, sizeof(Uint32), fp);
		if (feof(fp)) { THROW("Malformed 3D file"); }

		for(Uint32 i=0; i<noOfTris; i++) {
		    coord r,g,b;
		    fread(&idx1,1,4,fp);
		    if (feof(fp)) { THROW("Malformed 3D file (idx1)"); }
		    fread(&idx2,1,4,fp);
		    if (feof(fp)) { THROW("Malformed 3D file (idx2)"); }
		    fread(&idx3,1,4,fp);
		    if (feof(fp)) { THROW("Malformed 3D file (idx3)"); }

		    if (magic == TRI_MAGIC || magic == TRI_MAGICNORMAL) {
			fread(&r,1,4,fp);
			if (feof(fp)) { THROW("Malformed 3D file (r)"); }
			fread(&g,1,4,fp);
			if (feof(fp)) { THROW("Malformed 3D file (g)"); }
			fread(&b,1,4,fp);
			if (feof(fp)) { THROW("Malformed 3D file (b)"); }
		    } else {
			r = g = b = 1.0;
		    }
		    _triangles.push_back(
			new Triangle(
			    _vertexes[idx1],
			    _vertexes[idx2],
			    _vertexes[idx3], r,g,b));
		}

		totalPoints += noOfPoints;
		totalTris   += noOfTris;

		if (g_verbose>0) {
		    cout << "Vertexes: " << totalPoints;
		    cout << " Triangles: " << totalTris << endl;
		}
	    } while(!feof(fp));

	    fclose(fp);
	    _vertexes.reserve(totalTris*3);

	    if (magic != TRI_MAGICNORMAL)
		fixNormals();
	} else
	    THROW("Unknown extension (only .tri and .ply accepted)");
    }
    catch(RendererException *ex)
    {
	ex->DebugInfo();
	exit(1);
    }
}

void Object::fixNormals()
{
    std::vector<Vertex*>::iterator vertex;
    for(vertex = _vertexes.begin(); vertex != _vertexes.end(); vertex++)  {
        (*vertex)->getNormal()->set(0.,0.,0.);

        std::vector<Normal*>::iterator nrm;
        for(nrm = (*vertex)->getNormals().begin(); nrm != (*vertex)->getNormals().end(); nrm++)
            delete (*nrm);
        (*vertex)->getNormals().clear();
    }

    // Calculate normals
    coord aax,aay,aaz, bbx,bby,bbz, dumx,dumy,dumz;
    std::vector<Triangle*>::iterator tri;
    for(tri = _triangles.begin(); tri!=_triangles.end();) {

        Point* A = (*tri)->getV1();
        Point* B = (*tri)->getV2();
        Point* C = (*tri)->getV3();

        aax = B->getX() - A->getX();
        aay = B->getY() - A->getY();
        aaz = B->getZ() - A->getZ();
        coord norm = sqrt(aax*aax+aay*aay+aaz*aaz);
        aax/=norm;
        aay/=norm;
        aaz/=norm;

        bbx = C->getX() - B->getX();
        bby = C->getY() - B->getY();
        bbz = C->getZ() - B->getZ();
        norm = sqrt(bbx*bbx+bby*bby+bbz*bbz);
        bbx/=norm;
        bby/=norm;
        bbz/=norm;

        dumx=aay*bbz-bby*aaz;
        dumy=bbx*aaz-aax*bbz;
        dumz=aax*bby-aay*bbx;
        coord summ = sqrt(dumx*dumx+dumy*dumy+dumz*dumz);

        if (summ>1e-2) {
            dumx = dumx/summ;
            dumy = dumy/summ;
            dumz = dumz/summ;
        } else {
            if (g_verbose > 1) {
                cerr << "A:" << *A << endl;
                cerr << "B:" << *B << endl;
                cerr << "C:" << *C << endl;
                cerr << "These don't define a nice triangle... :(\n";
            }
            //THROW("calculation of normals");
            tri = _triangles.erase(tri);
            continue;
        }

        Normal *nrm1 = new Normal(dumx, dumy, dumz, &((*tri)->_v[0]));
        (*tri)->getV1()->addNormal(nrm1);
        Normal *nrm2 = new Normal(dumx, dumy, dumz, &((*tri)->_v[1]));
        (*tri)->getV2()->addNormal(nrm2);
        Normal *nrm3 = new Normal(dumx, dumy, dumz, &((*tri)->_v[2]));
        (*tri)->getV3()->addNormal(nrm3);

        (*tri)->getNormal().set(dumx,dumy,dumz);
        (*tri)->updateCenter();
        tri++;
    }

    for(vertex = _vertexes.begin(); vertex != _vertexes.end(); )
    {
        switch ((*vertex)->findBreakingNormal(this)) {
        case -1:
            vertex = _vertexes.erase(vertex);
            break;
        case 0:
            vertex++;
            break;
        case 1:
        default:
            vertex++;
        }
    }
}

void Object::renderPoints(Camera& camera)
{
    glTranslatef(_x, _y, _z);
    glRotatef(_rx, 1.,0.,0.);
    glRotatef(_ry, 0.,1.,0.);
    glRotatef(_rz, 0.,0.,1.);
    if (g_VBOEnabled) {
	glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[0] );
	glVertexPointer( 3, GL_FLOAT, 0, (char *) NULL);
	glEnableClientState( GL_VERTEX_ARRAY );
	glDrawArrays( GL_POINTS, 0, _vertexes.size() );
	glDisableClientState( GL_VERTEX_ARRAY );
    } else {
	std::vector<Vertex*>::iterator vertex;
	for(vertex = _vertexes.begin();
		vertex != _vertexes.end();
		vertex++)
	{
	    camera.getScreen()->plot(*vertex);
	}
    }
}

void Object::renderGouraud(Camera& camera)
{
    glTranslatef(_x, _y, _z);
    glRotatef(_rx, 1.,0.,0.);
    glRotatef(_ry, 0.,1.,0.);
    glRotatef(_rz, 0.,0.,1.);
    std::vector<Triangle*>::iterator tri;
    for(tri = _triangles.begin(); tri!=_triangles.end(); tri++) {
        camera.getScreen()->drawTriangleGouraud(*tri);
    }
}

void Object::renderPhong(Camera& camera)
{
    glTranslatef(_x, _y, _z);
    glRotatef(_rx, 1.,0.,0.);
    glRotatef(_ry, 0.,1.,0.);
    glRotatef(_rz, 0.,0.,1.);
    if (g_VBOEnabled) {
	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_NORMAL_ARRAY );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );
	glEnableClientState( GL_COLOR_ARRAY );

	glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[1] );
	glVertexPointer( 3, GL_FLOAT, 0, (char *) NULL);
	glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[2] );
	glNormalPointer( GL_FLOAT, 0, (char *) NULL);
	glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[3] );
	glTexCoordPointer( 3, GL_FLOAT, 0, (char *) NULL);
	glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[4] );
	glColorPointer( 3, GL_FLOAT, 0, (char *) NULL);

	glDrawArrays( GL_TRIANGLES, 0, _triangles.size()*3 );

	glDisableClientState( GL_COLOR_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_VERTEX_ARRAY );
    } else {
	std::vector<Triangle*>::iterator tri;
	for(tri = _triangles.begin(); tri!=_triangles.end(); tri++) {
	    camera.getScreen()->drawTrianglePhong(*tri);
	}
    }
}

void CopyGLTripletToFloatPointer(float* &pDest, float *pSrc)
{
    *pDest++ = pSrc[0];
    *pDest++ = pSrc[1];
    *pDest++ = pSrc[2];
}

void Object::PrepareVBO()
{
    if (!g_VBOEnabled)
	return;

    glGenBuffersARB(5, _VBObuffers);
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }

    float *pVBOvertexCache = new float[_vertexes.size()*3];
    if (!pVBOvertexCache)
	THROW("Out of memory");
    float *pVertexData = pVBOvertexCache;

    // First VBO buffer: vertexes for point rendering
    std::vector<Vertex*>::iterator vrt;
    for(vrt = _vertexes.begin(); vrt!=_vertexes.end(); vrt++) 
	CopyGLTripletToFloatPointer(pVertexData, (*vrt)->getArrayGL());

    glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[0] );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBufferDataARB( GL_ARRAY_BUFFER_ARB, _vertexes.size()*3*sizeof(float), pVBOvertexCache, GL_STATIC_DRAW_ARB );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }

    delete [] pVBOvertexCache;

    // Rest VBO buffers: triangle rendering, (vertexes, normals, ambOcl(textures), colors)
    float *pVBOTriVertexCache = new float[_triangles.size()*3*3];
    if (!pVBOTriVertexCache) THROW("Out of memory");
    float *pVBOTriNormalCache = new float[_triangles.size()*3*3];
    if (!pVBOTriNormalCache) THROW("Out of memory");
    float *pVBOTriTextureCache = new float[_triangles.size()*3*3];
    if (!pVBOTriTextureCache) THROW("Out of memory");
    float *pVBOTriColorCache = new float[_triangles.size()*3*3];
    if (!pVBOTriColorCache) THROW("Out of memory");

    float *pTriVertexData = pVBOTriVertexCache;
    float *pTriNormalData = pVBOTriNormalCache;
    float *pTriTextureData = pVBOTriTextureCache;
    float *pTriColorData = pVBOTriColorCache;
    std::vector<Triangle*>::iterator tri;
    for(tri = _triangles.begin(); tri!=_triangles.end(); tri++) {
        Triangle *t = *tri;
	CopyGLTripletToFloatPointer(pTriVertexData, t->getV1()->getArrayGL());
	CopyGLTripletToFloatPointer(pTriVertexData, t->getV2()->getArrayGL());
	CopyGLTripletToFloatPointer(pTriVertexData, t->getV3()->getArrayGL());

	CopyGLTripletToFloatPointer(pTriNormalData, t->getV1()->getNormal()->getArrayGL());
	CopyGLTripletToFloatPointer(pTriNormalData, t->getV2()->getNormal()->getArrayGL());
	CopyGLTripletToFloatPointer(pTriNormalData, t->getV3()->getNormal()->getArrayGL());

	CopyGLTripletToFloatPointer(pTriTextureData, t->getV1()->getAmbientOcclusionFactorArrayGL());
	CopyGLTripletToFloatPointer(pTriTextureData, t->getV2()->getAmbientOcclusionFactorArrayGL());
	CopyGLTripletToFloatPointer(pTriTextureData, t->getV3()->getAmbientOcclusionFactorArrayGL());

	CopyGLTripletToFloatPointer(pTriColorData, t->getColorGL());
	CopyGLTripletToFloatPointer(pTriColorData, t->getColorGL());
	CopyGLTripletToFloatPointer(pTriColorData, t->getColorGL());
    }
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[1] );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBufferDataARB( GL_ARRAY_BUFFER_ARB, _triangles.size()*3*3*sizeof(float), pVBOTriVertexCache, GL_STATIC_DRAW_ARB );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[2] );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBufferDataARB( GL_ARRAY_BUFFER_ARB, _triangles.size()*3*3*sizeof(float), pVBOTriNormalCache, GL_STATIC_DRAW_ARB );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[3] );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBufferDataARB( GL_ARRAY_BUFFER_ARB, _triangles.size()*3*3*sizeof(float), pVBOTriTextureCache, GL_STATIC_DRAW_ARB );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBindBufferARB( GL_ARRAY_BUFFER_ARB, _VBObuffers[4] );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }
    glBufferDataARB( GL_ARRAY_BUFFER_ARB, _triangles.size()*3*3*sizeof(float), pVBOTriColorCache, GL_STATIC_DRAW_ARB );
    if (GL_NO_ERROR != glGetError()) { g_VBOEnabled = false; return; }

    delete [] pVBOTriVertexCache;
    delete [] pVBOTriNormalCache;
    delete [] pVBOTriTextureCache;
    delete [] pVBOTriColorCache;
}

void Object::stretch(coord factor)
{
    _maxx = _maxy = _maxz = -1.e10;
    _minx = _miny = _minz = 1.e10;

    std::vector<Vertex*>::iterator vertex;
    for(vertex = _vertexes.begin();
            vertex != _vertexes.end();
            vertex++)
    {
        (*vertex)->set(
            factor * (*vertex)->getX(),
            factor * (*vertex)->getY(),
            factor * (*vertex)->getZ());
    }

    fixNormals();
}

void Object::save()
{
    FILE *fp = fopen("conv.tri" ,"wb");
    float ftmp;
    unsigned tmp = TRI_MAGICNORMAL;
    fwrite(&tmp, 1, 4, fp);
    tmp = _vertexes.size(); fwrite(&tmp, 1, 4, fp);
    std::vector<Vertex*>::iterator it = _vertexes.begin();
    while(it != _vertexes.end()) {
        ftmp = (*it)->getX(); fwrite(&ftmp, 1, 4, fp);
        ftmp = (*it)->getY(); fwrite(&ftmp, 1, 4, fp);
        ftmp = (*it)->getZ(); fwrite(&ftmp, 1, 4, fp);
        ftmp = (*it)->getNormal()->getX(); fwrite(&ftmp, 1, 4, fp);
        ftmp = (*it)->getNormal()->getY(); fwrite(&ftmp, 1, 4, fp);
        ftmp = (*it)->getNormal()->getZ(); fwrite(&ftmp, 1, 4, fp);
        it++;
    }
    tmp = _triangles.size(); fwrite(&tmp, 1, 4, fp);
    std::vector<Triangle*>::iterator tri;
    for(tri = _triangles.begin(); tri!=_triangles.end(); tri++) {
        tmp = find(_vertexes.begin(), _vertexes.end(), (*tri)->getV1()) - _vertexes.begin();
        assert(tmp>=0);
        fwrite(&tmp, 1, 4, fp);
        tmp = find(_vertexes.begin(), _vertexes.end(), (*tri)->getV2()) - _vertexes.begin();
        assert(tmp>=0);
        fwrite(&tmp, 1, 4, fp);
        tmp = find(_vertexes.begin(), _vertexes.end(), (*tri)->getV3()) - _vertexes.begin();
        assert(tmp>=0);
        fwrite(&tmp, 1, 4, fp);

        GLfloat *ftmp = (*tri)->getColorGL();
        fwrite(&ftmp[0], 1, 4, fp);
        fwrite(&ftmp[1], 1, 4, fp);
        fwrite(&ftmp[2], 1, 4, fp);
    }

    fclose(fp);
}
