/*
 *  TriMeshOpenGLViewer-Phong
 *  Copyright (C) 2005  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define GL_GLEXT_PROTOTYPES

#include <GL/gl.h>
#include <GL/glext.h>

#include <assert.h>

#include "Scene.h"
#include "Object.h"
#include "Variables.h"

void Scene::addLight(Light* light) 
{
    _lights.push_back(light);
    light->setScene(this);
}

void Scene::addCamera(Camera* camera) 
{
    _cameras.push_back(camera);
}

void Scene::addObject(Object *object) 
{
    _objects.push_back(object);
    object->setScene(this);
}

typedef struct {
    std::vector<Light*>::iterator light;
    coord distance;
} LightSorter;

void Scene::DrawLights(Camera& camera)
{
    glDisable(GL_LIGHTING);
    glDisable(GL_CULL_FACE);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, camera.getScreen()->getFirstTexture());
    std::vector<Light*>::iterator l;
    l = _lights.begin();

    // Sort lights by distance (since only 8 are allowed, bubble-sort)
    LightSorter sorter[8];
    int noLights = 0;
    while(l != _lights.end()) {
        sorter[noLights  ].light = l;
        sorter[noLights++].distance =
            (camera.getX()-(*l)->getX())*(camera.getX()-(*l)->getX())+
            (camera.getY()-(*l)->getY())*(camera.getY()-(*l)->getY())+
            (camera.getZ()-(*l)->getZ())*(camera.getZ()-(*l)->getZ());
        l++;
    }
    int i;
    for (i=0; i<noLights; i++)
        for(int j=i+1; j<noLights; j++)
            if (sorter[i].distance < sorter[j].distance) {
                LightSorter temp = sorter[i];
                sorter[i] = sorter[j];
                sorter[j] = temp;
            }

    for (i=0; i<noLights; i++) {
        glLoadIdentity();
        camera.lookGLU();

        (*sorter[i].light)->DrawLightUsingBitmapSun(camera);
    }
}

template <typename ObjectWorker, typename LightsWorker>
struct SceneWork {
    static void CallMethodForAllObjects(Scene& scene, Camera& camera) {
	std::vector<Object*>::iterator obj;
	obj = scene._objects.begin();
	while(obj != scene._objects.end()) {

	    glLoadIdentity();
	    camera.lookGLU();

	    LightsWorker::DoWork(scene, camera);

	    ObjectWorker::DoWork(scene, camera, *obj);
	    obj++;
	}
    }
};

struct NoLightWork {
    static void DoWork(Scene& scene, Camera& camera) {}
};

struct InvokeRenderPoints {
    static void DoWork(Scene& scene, Camera& camera, Object *pObj) {
	pObj->renderPoints(camera);
    }
};
    
void Scene::renderPoints(Camera& camera)
{
    SceneWork<InvokeRenderPoints, NoLightWork>::CallMethodForAllObjects(*this, camera);
}

struct SetupLightsInOpenGL {
    static void DoWork(Scene& scene, Camera& camera) {
	std::vector<Light*>::iterator light = scene._lights.begin();
	while(light != scene._lights.end()) {
	    (*light)->SetupLightInOpenGL(camera);
	    light++;
	}
    }
};
    
struct InvokeRenderGouraud {
    static void DoWork(Scene& scene, Camera& camera, Object *pObj) {
	pObj->renderGouraud(camera);
    }
};

void Scene::renderGouraud(Camera& camera, bool isWireframe)
{
    if (isWireframe)
        glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);

    glEnable(GL_LIGHTING);
    g_CullingEnabled?glEnable(GL_CULL_FACE):glDisable(GL_CULL_FACE);
    glDisable(GL_TEXTURE_2D);

    SceneWork<InvokeRenderGouraud, SetupLightsInOpenGL>::CallMethodForAllObjects(*this, camera);

    if (isWireframe)
        glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);

    DrawLights(camera);
}

struct InvokeRenderPhong {
    static void DoWork(Scene& s, Camera& camera, Object *pObj) {
	assert(s._lights.begin() != s._lights.end());
	Light *pFirstLight = (*s._lights.begin());
	GLfloat vLight[4] = { pFirstLight->getX(), pFirstLight->getY(), pFirstLight->getZ(), 1.0 };
	glProgramEnvParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 1, vLight);
	GLfloat *pEye = camera.getArrayGL();
	glProgramEnvParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 2, pEye);
        pObj->renderPhong(camera);
    }
};

void Scene::renderPhong(Camera& camera)
{
    if (!g_PhongEnabled)
	return;

    if (_lights.size() != 1) {
	cerr << "Phong fragment shader only works with one light!" << endl;
	g_PhongEnabled = false;
	return;
    }
    glEnable(GL_LIGHTING);
    glEnable(GL_VERTEX_PROGRAM_ARB);
    glEnable(GL_FRAGMENT_PROGRAM_ARB);
    g_CullingEnabled?glEnable(GL_CULL_FACE):glDisable(GL_CULL_FACE);
    glDisable(GL_TEXTURE_2D);

    SceneWork<InvokeRenderPhong, SetupLightsInOpenGL>::CallMethodForAllObjects(*this, camera);

    glDisable(GL_VERTEX_PROGRAM_ARB);
    glDisable(GL_FRAGMENT_PROGRAM_ARB);

    DrawLights(camera);
}
