/*
 *  TriMeshOpenGLViewer-Phong
 *  Copyright (C) 2005  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <string.h>
#include <math.h>
#include <sys/stat.h>

#define GL_GLEXT_PROTOTYPES

#include <GL/gl.h>
#include <GL/glext.h>
#include <GL/glu.h>

#include "Screen3D.h"
#include "Scene.h"
#include "Exceptions.h"
#include "Vertex.h"
#include "Variables.h"
#include "Triangle.h"

#include "lightImage.h"

#define NProgs		2
static GLuint		Shader[NProgs];

static const char * ShaderSrc[NProgs];


static void getErrorLine (char * src, int err, char line[], int maxLine)
{
    int start, finish, n;

    start = err;
    while (start >= 0 && isprint(src[start]))
        start --;
    start ++;
    finish = err;
    while (isprint(src[finish]))
        finish ++;
    n = finish - start;
    if (n >= maxLine)
        n = maxLine;
    strncpy(line, src + start, n);
    line[n] = 0;
}

static void checkShader (int i)
{
    int  err;
    char msg[256];

    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &err);
    if (err >= 0) {
        getErrorLine(const_cast<char *>(ShaderSrc[i]), err, msg, sizeof(msg));
        fprintf(stderr, "Error in shader #%d: %s\n", i,
               glGetString(GL_PROGRAM_ERROR_STRING_ARB));
        fprintf(stderr, "Source line:    %s\n", msg);
	fflush(stderr);
        exit(-1);
    }
}

const char *vertexProgram = \
    "!!ARBvp1.0\n"
    "\n"
    "OPTION ARB_position_invariant;\n"
    "\n"
    "PARAM lightPos  = program.env[1];\n"
    "PARAM cameraPos = program.env[2];\n"
    "PARAM m[4]={state.matrix.modelview};\n"
    "PARAM mvinv[4]={state.matrix.modelview.invtrans};\n"
    "PARAM zero = 0.0;\n"
    "\n"
    "TEMP EyeCamPos, EyeNormal, EyeVertex;\n"
    "\n"
    "# Vertex position in eye space\n"
    "\n"
    "DP4     EyeVertex.x, m[0], vertex.position;\n"
    "DP4     EyeVertex.y, m[1], vertex.position;\n"
    "DP4     EyeVertex.z, m[2], vertex.position;\n"
    "\n"
    "# Camera pos in eye space.\n"
    "\n"
    "DP4     EyeCamPos.x, m[0], cameraPos;\n"
    "DP4     EyeCamPos.y, m[1], cameraPos;\n"
    "DP4     EyeCamPos.z, m[2], cameraPos;\n"
    "\n"
    "# Normal vector in eye space.\n"
    "\n"
    "DP3     EyeNormal.x, mvinv[0], vertex.normal;\n"
    "DP3     EyeNormal.y, mvinv[1], vertex.normal;\n"
    "DP3     EyeNormal.z, mvinv[2], vertex.normal;\n"
    "\n"
    "# Normalize the normal :)\n"
    "\n"
    "DP3 EyeNormal.w, EyeNormal, EyeNormal;\n"
    "RSQ EyeNormal.w, EyeNormal.w;\n"
    "MUL EyeNormal, EyeNormal, EyeNormal.w;\n"
    "\n"
    "# Set the values to interpolate over each pixel:\n"
    "# 1. The vector (in eye space) from the vertex to the light\n"
    "# 2. The vector (in eye space) from the vertex to the camera\n"
    "# 3. The normal vector\n"
    "\n"
    "SUB     result.texcoord[1], state.light[0].position, EyeVertex;\n"
    "SUB     result.texcoord[2], EyeCamPos, EyeVertex;\n"
    "MOV     result.texcoord[3], EyeNormal;\n"
    "MOV     result.texcoord[4], vertex.texcoord[0];\n"
    "\n"
    "# Pass the vertex color to the fragment shader (for lighting)\n"
    "\n"
    "MOV	result.color, vertex.color;\n"
    "\n"
    "END\n";

const char *fragmentProgram = \
    "!!ARBfp1.0\n"
    "TEMP lightVec, viewVec, reflVec, normal, Len, finalCol, ambOclFactor, lightContrib;\n"
    "PARAM quarter  = {0.25, 0.25, 0.25, 0.25};\n"
    "PARAM ambientScale  = {0.375, 0.375, 0.375, 0.375};\n"
    "PARAM half  = {0.5, 0.5, 0.5, 0.5};\n"
    "PARAM nova  = {1.0, 1.0, 1.0, 0.0};      # Specular color\n"
    "PARAM spec_coeff = 64.0;                 # Specular coefficient\n"
    "#PARAM two   = {2.0, 2.0, 2.0, 2.0};     # Used in Phong lighting\n"
    "\n"
    "# Normalize interpolated normal, lightvec and viewvec.\n"
    "\n"
    "DP3             Len.w, fragment.texcoord[1], fragment.texcoord[1];\n"
    "RSQ             lightVec.w, Len.w;\n"
    "MUL             lightVec.xyz, fragment.texcoord[1], lightVec.w;\n"
    "\n"
    "DP3             viewVec.w, fragment.texcoord[2], fragment.texcoord[2];\n"
    "RSQ             viewVec.w, viewVec.w;\n"
    "MUL             viewVec.xyz, fragment.texcoord[2], viewVec.w;\n"
    "\n"
    "DP3             normal.w, fragment.texcoord[3], fragment.texcoord[3];\n"
    "RSQ             normal.w, normal.w;\n"
    "MUL             normal.xyz, fragment.texcoord[3], normal.w;\n"
    "\n"
    "# Diffuse contribution\n"
    "\n"
    "DP3_SAT lightContrib.x, normal, lightVec;\n"
    "\n"
    "# Specular contribution\n"
    "\n"
    "# Phong lighting (enable it for smaller highlights)\n"
    "#DP3             reflVec, lightVec, normal;\n"
    "#MUL             reflVec, reflVec, two;\n"
    "#MAD             reflVec, reflVec, normal, -lightVec;\n"
    "#DP3_SAT lightContrib.y, reflVec, viewVec;\n"
    "\n"
    "# Blinn lighting:\n"
    "ADD             reflVec, lightVec, viewVec;\n"
    "DP3             reflVec.w, reflVec, reflVec;\n"
    "RSQ             reflVec.w, reflVec.w;\n"
    "MUL             reflVec.xyz, reflVec, reflVec.w;\n"
    "DP3             lightContrib.y, reflVec, normal;\n"
    "\n"
    "MOV             lightContrib.w, spec_coeff; # specular coefficient\n"
    "\n"
    "# Lighting computations\n"
    "\n"
    "LIT_SAT lightContrib, lightContrib;\n"
    "\n"
    "# total light calculated per pixel by this formula:\n"
    "# DIFFUSE*diffuse + AMBIENT*ambientOcclusion + specular\n"
    "\n"
    "# DIFFUSE = 128/255 = 0.5, scale the diffuse factor by the 'half' PARAM\n"
    "MUL             lightContrib.y, lightContrib.y, half;\n"   
    "\n"
    "# DIFFUSE*diffuse*fragment.color\n"
    "MUL             finalCol, lightContrib.y, fragment.color;\n"
    "\n"
    "# add nova*specular (nova=1.0)\n"
    "MAD             finalCol, lightContrib.z, nova, finalCol;\n"
    "\n"
    "# AMBIENT = 96/255 = 0.375, scale the interpolated ambientOcclusion by ambientScale\n"
    "MUL             ambOclFactor.x, fragment.texcoord[4].x, ambientScale.x;\n"   
    "\n"
    "# add ambientOcclusion*fragment.color\n"
    "MAD             finalCol, ambOclFactor.x, fragment.color, finalCol;\n"
    "MOV             result.color.xyz, finalCol;\n"
    "MOV             result.color.w, state.material.diffuse.w;\n"
    "\n"
    "END\n";

void Screen3D::initShader()
{
    glEnable(GL_VERTEX_PROGRAM_ARB);
    if (GL_NO_ERROR != glGetError()) { g_PhongEnabled = false; return; }

    glEnable(GL_FRAGMENT_PROGRAM_ARB);
    if (GL_NO_ERROR != glGetError()) { g_PhongEnabled = false; return; }

    glGenProgramsARB(NProgs, Shader);
    if (GL_NO_ERROR != glGetError()) { g_PhongEnabled = false; return; }

    glBindProgramARB(GL_VERTEX_PROGRAM_ARB, Shader[0]);
    if (GL_NO_ERROR != glGetError()) { g_PhongEnabled = false; return; }

    ShaderSrc[0] = vertexProgram;
    ShaderSrc[1] = fragmentProgram;

    glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB,
                       strlen(ShaderSrc[0]), ShaderSrc[0]);
    if (GL_NO_ERROR != glGetError()) { g_PhongEnabled = false; return; }

    checkShader(0);

    glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, Shader[1]);
    if (GL_NO_ERROR != glGetError()) { g_PhongEnabled = false; return; }

    glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB,
                       strlen(ShaderSrc[1]), ShaderSrc[1]);
    if (GL_NO_ERROR != glGetError()) { g_PhongEnabled = false; return; }

    checkShader(1);

    glDisable(GL_VERTEX_PROGRAM_ARB);
    glDisable(GL_FRAGMENT_PROGRAM_ARB);
}

Screen3D::Screen3D()
{
    if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
        cerr << "Couldn't initialize SDL: " <<  SDL_GetError() << endl;
        THROW("Screen3D");
    }

    // Clean up on exit
    atexit(SDL_Quit);

    if (!SDL_SetVideoMode(
                WIDTH,
                HEIGHT,
                0,
                SDL_OPENGL))
    {
        cerr << "Couldn't set video mode: " << SDL_GetError() << endl;
        THROW("Screen3D");
    }

    InitGL();
    loadGLTexture();

    char windowTitle[256];
    strcpy(windowTitle, "final3D");
    setTitle(windowTitle);
}

void Screen3D::InitGL()
{
    glViewport(0, 0, WIDTH, HEIGHT);
    glClearColor(0.3f, 0.3f, 0.3f, 0.5f);         // This Will Clear The Background Color To Black
    glClearDepth(1.0);                            // Enables Clearing Of The Depth Buffer
    glEnable(GL_DEPTH_TEST);                      // Enables Depth Testing
    glDepthFunc(GL_LESS);                         // The Type Of Depth Test To Do
    glShadeModel(GL_SMOOTH);                      // Enables Smooth Color Shading
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();                             // Reset The Projection Matrix
    gluPerspective(45.0f,(GLfloat)WIDTH/(GLfloat)HEIGHT,0.1f,100.0f);     // Calculate The Aspect Ratio Of The Window
    glMatrixMode(GL_MODELVIEW);
}

void Screen3D::setTitle(const char *title)
{
    SDL_WM_SetCaption(title, "final3D");
    SDL_EnableUNICODE(1);
}

Screen3D::~Screen3D()
{
}

void Screen3D::clearPage()
{
    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
    glLoadIdentity();
}

void Screen3D::showPage()
{
    SDL_GL_SwapBuffers();
}

void Screen3D::plot(Vertex *a)
{
    glBegin(GL_POINTS);
    glColor4f(1.,1.,1.,1.);
    glVertex3fv( a->getArrayGL() );
    glEnd();
}

void Screen3D::drawTriangleGouraud(Triangle *t)
{
    glMaterialfv(GL_FRONT, GL_AMBIENT, t->getAmbientOcclusionColorGL());
    glMaterialfv(GL_FRONT, GL_DIFFUSE, t->getColorGL());
    glBegin(GL_TRIANGLES);
    glNormal3fv( t->getV1()->getNormal()->getArrayGL() );
    glVertex3fv( t->getV1()->getArrayGL() );
    glNormal3fv( t->getV2()->getNormal()->getArrayGL() );
    glVertex3fv( t->getV2()->getArrayGL() );
    glNormal3fv( t->getV3()->getNormal()->getArrayGL() );
    glVertex3fv( t->getV3()->getArrayGL() );
    glEnd();
}

void Screen3D::drawTrianglePhong(Triangle *t)
{
    glColor3fv( t->getColorGL());
    glBegin(GL_TRIANGLES);
    glTexCoord3fv(t->getV1()->getAmbientOcclusionFactorArrayGL());
    glNormal3fv( t->getV1()->getNormal()->getArrayGL() );
    glVertex3fv( t->getV1()->getArrayGL() );
    glTexCoord3fv(t->getV2()->getAmbientOcclusionFactorArrayGL());
    glNormal3fv( t->getV2()->getNormal()->getArrayGL() );
    glVertex3fv( t->getV2()->getArrayGL() );
    glTexCoord3fv(t->getV3()->getAmbientOcclusionFactorArrayGL());
    glNormal3fv( t->getV3()->getNormal()->getArrayGL() );
    glVertex3fv( t->getV3()->getArrayGL() );
    glEnd();
}

void Screen3D::loadGLTexture()
{
    static Uint8 *lightTMap = (Uint8 *) &lightPPM[15];

    // Load light texture map
    glGenTextures(1, &_textures[0]);
    glBindTexture(GL_TEXTURE_2D, _textures[0]);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexImage2D( GL_TEXTURE_2D, 0, 3, 128, 128, 0, GL_RGB, GL_UNSIGNED_BYTE, lightTMap);
}
