/*
 *  TriMeshOpenGLViewer-Phong
 *  Copyright (C) 2005  Thanassis Tsiodras (ttsiod@softlab.ntua.gr)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef _WIN32
#include <windows.h>
#else
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glxext.h>
#endif

#include <math.h>

#define GLOBAL_DEFS
#include "Variables.h"

#include "Screen3D.h"
#include "Scene.h"
#include "Object.h"
#include "Camera.h"
#include "Keyboard.h"
#include "Clock.h"
#include "Vertex.h"
#include "Triangle.h"

#define DEG2RAD(x) ((x)*M_PI/180.)

int main(int argc, char *argv[])
{
    coord angle3=120.0;
    coord angle4=0.0*M_PI/180.;
    Keyboard keys;
    coord ex,ey,ez, lookleft, lookup;
    RenderState drawingMode = RENDER_PHONG;
    Object *obj;
    bool autospin = true;

    switch (argc)  {
    case 2:
        obj = new Object(argv[1]);
        break;
    case 3:
        if (!strcmp(argv[1], "-nocull")) {
            g_CullingEnabled = false;
            obj = new Object(argv[2]);
        }
        else {
            THROW("No such option\nUsage: final3d [-nocull] file.tri\n");
        }
        break;
    default:
    { THROW("No such options (too many?)\nUsage: final3d [-nocull] file.tri\n"); }
    }
    coord wx = obj->getMaxX()-obj->getMinX();
    coord wy = obj->getMaxY()-obj->getMinY();
    coord wz = obj->getMaxZ()-obj->getMinZ();
    if (wx<wy) wx=wy;
    if (wx<wz) wx=wz;
    obj->stretch(1.0/wx);
    wx = obj->getMaxX()+obj->getMinX();
    wy = obj->getMaxY()+obj->getMinY();
    wz = obj->getMaxZ()+obj->getMinZ();
    obj->place(-wx/2, -wy/2, -wz/2);

    Screen3D canvas;

    const char *extensions = (const char *) glGetString(GL_EXTENSIONS);
    //puts(extensions);
    if (!strstr(extensions, "GL_ARB_vertex_program") ||
	!strstr(extensions, "GL_ARB_fragment_program"))
    {
	cerr << "Missing one of:\n";
	cerr << "\tGL_ARB_vertex_program\n";
	cerr << "\tGL_ARB_fragment_program\n";
	cerr << "Phong mode disabled... :-(" << endl;
	g_PhongEnabled = false;
	drawingMode = RENDER_GOURAUD;
    } else {
	g_PhongEnabled = true;
    }

    if (!strstr(extensions, "GL_ARB_vertex_buffer_object"))
	g_VBOEnabled = false;
    else
	g_VBOEnabled = true;

#ifdef _WIN32
    glProgramEnvParameter4fvARB = (GlProgramEnvParameter4fvARB) wglGetProcAddress("glProgramEnvParameter4fvARB");
    glGenProgramsARB = (GlGenProgramsARB) wglGetProcAddress("glGenProgramsARB");
    glBindProgramARB = (GlBindProgramARB) wglGetProcAddress("glBindProgramARB");
    glProgramStringARB = (GlProgramStringARB) wglGetProcAddress("glProgramStringARB");
    if (!glProgramEnvParameter4fvARB || 
	!glGenProgramsARB ||
	!glBindProgramARB ||
	!glProgramStringARB) 
    {
	g_PhongEnabled = false;
    }
    glGenBuffersARB = (GlGenBuffersARB) wglGetProcAddress("glGenBuffersARB");
    glBindBufferARB = (GlBindBufferARB) wglGetProcAddress("glBindBufferARB");
    glBufferDataARB = (GlBufferDataARB) wglGetProcAddress("glBufferDataARB");
    if (!glGenBuffersARB || 
	!glBindBufferARB ||
	!glBufferDataARB) 
    {
	g_VBOEnabled = false;
    }
#endif

    try {
	if (g_PhongEnabled) 
	    canvas.initShader();
    }
    catch(...) {
	g_PhongEnabled = false;
    }

    if (!g_PhongEnabled)
	cerr << "Phong extensions (ProgramsARB) are not available, phong mode disabled...\n";

    if (g_VBOEnabled)
	obj->PrepareVBO();

    Scene scene1;
    scene1.addObject(obj);

    ex = 2.0;
    ey = 0.0;
    ez = 0.0;
    lookleft = 0.;
    lookup = 0.;
    Camera sony(ex,ey,ez, lookleft, lookup, &scene1, &canvas);

    coord radiusl = 1.7;
    Light torch(0.0, 0.0, 3.0); // High above, on the Z-axis
    scene1.addLight(&torch);
    keys.poll();

    Uint32 framesDrawn = 0;
    Clock watch;

    sony.setRenderState(drawingMode);
    switch(drawingMode) {
    case RENDER_POINTS:
        canvas.setTitle("Point rendering"); break;
    case RENDER_LINES:
        canvas.setTitle("AA lines rendering"); break;
    case RENDER_GOURAUD:
        canvas.setTitle("ZBuffer,Gouraud rendering");break;
    case RENDER_PHONG:
        canvas.setTitle("ZBuffer,Phong rendering"); break;
    }
    canvas.showPage();
    canvas.clearPage();


    coord dAngle = 0.3f;
    coord dMove = dAngle/30.;

    while(!keys.Abort()) {
        framesDrawn++;
        canvas.showPage();
        canvas.clearPage();
        if (keys.Left())
            lookleft+=dAngle;
        if (keys.Right())
            lookleft-=dAngle;
        if (keys.Up()) {
            lookup+=dAngle; if (lookup>89.) lookup = 89.;
        }
        if (keys.Down()) {
            lookup-=dAngle; if (lookup<-89.) lookup = -89.;
        }
        if (keys.Forward()) {
            ex -= dMove*cos(DEG2RAD(lookup))*cos(DEG2RAD(lookleft));
            ey -= dMove*cos(DEG2RAD(lookup))*sin(DEG2RAD(lookleft));
            ez += dMove*sin(DEG2RAD(lookup));
        }
        if (keys.Backward()) {
            ex += dMove*cos(DEG2RAD(lookup))*cos(DEG2RAD(lookleft));
            ey += dMove*cos(DEG2RAD(lookup))*sin(DEG2RAD(lookleft));
            ez -= dMove*sin(DEG2RAD(lookup));
        }
	if (keys.isS()) {
            ex -= dMove*cos(DEG2RAD(0))*cos(DEG2RAD(lookleft+90.));
            ey -= dMove*cos(DEG2RAD(0))*sin(DEG2RAD(lookleft+90.));
            ez += dMove*sin(DEG2RAD(0));
	}
	if (keys.isF()) {
            ex -= dMove*cos(DEG2RAD(0))*cos(DEG2RAD(lookleft-90.));
            ey -= dMove*cos(DEG2RAD(0))*sin(DEG2RAD(lookleft-90.));
            ez += dMove*sin(DEG2RAD(0));
	}
        if (keys.isE()) {
            ex -= dMove*cos(DEG2RAD(lookup+90.))*cos(DEG2RAD(lookleft));
            ey -= dMove*cos(DEG2RAD(lookup+90.))*sin(DEG2RAD(lookleft));
            ez += dMove*sin(DEG2RAD(lookup+90.));
	}
        if (keys.isD()) {
            ex -= dMove*cos(DEG2RAD(lookup-90.))*cos(DEG2RAD(lookleft));
            ey -= dMove*cos(DEG2RAD(lookup-90.))*sin(DEG2RAD(lookleft));
            ez += dMove*sin(DEG2RAD(lookup-90.));
	}
	    // Automatically split shared normals (thus, make the tris appear separate)
	    //while(keys.isN()) keys.poll();
	    //g_interAngle-=5.;
	    //if (g_interAngle<5.)
	    //    g_interAngle = 5.;
	    //obj->fixNormals();
	    //obj->save();

        if (keys.isR()) {
            while(keys.isR()) keys.poll();
            autospin = !autospin;
        }
        if (keys.Light())
            angle3+=dAngle*4;
        if (keys.Mode()) {
            while(keys.Mode()) keys.poll();
            watch.reset();
            framesDrawn = 1;
            if (drawingMode == RENDER_POINTS)
                drawingMode = RENDER_LINES;
            else if (drawingMode == RENDER_LINES)
                drawingMode = RENDER_GOURAUD;
            else if (drawingMode == RENDER_GOURAUD)
                drawingMode = g_PhongEnabled?RENDER_PHONG:RENDER_POINTS;
            else if (drawingMode == RENDER_PHONG)
                drawingMode = RENDER_POINTS;
            sony.setRenderState(drawingMode);
            switch(drawingMode) {
            case RENDER_POINTS:
                canvas.setTitle("Point rendering"); break;
            case RENDER_LINES:
                canvas.setTitle("AA lines rendering"); break;
            case RENDER_GOURAUD:
                canvas.setTitle("ZBuffer,Gouraud rendering");break;
            case RENDER_PHONG:
                canvas.setTitle("ZBuffer,Phong rendering"); break;
            }
        }
        if (autospin) 
            angle4-=dAngle;
        obj->rotateZ(angle4);

        sony.set(ex,ey,ez, lookleft, lookup);
        torch.set( radiusl*cos(angle3*M_PI/180.0), radiusl*sin(angle3*M_PI/180.0), 0.8);
        sony.render();
        keys.poll();

	// The more frames per sec, the smaller the dAngle should be
	// The less frames per sec, the larger  the dAngle should be
	//   so...
	// Move dAngle towards 36/framesPerSec in 15 frames (steps)
	//   for 100fps, dAngle = 0.36
	//   for 10 fps, dAngle = 3.6
	// In both cases, we rotate 360 degress in 10 seconds
	//dAngle += ((36.0/(framesDrawn/(watch.readMS()/1000.0)))-dAngle)/15.;
	// 
	// Changed to 18/framesPerSec (20 seconds needed for a 360 rotation)
	dAngle += ((9.0/(framesDrawn/(watch.readMS()/1000.0)))-dAngle)/15.;
	dMove = dAngle/30.;
    }
    cout << "Rendering " << framesDrawn << " frames in ";
    cout << (watch.readMS()/1000.0) << " seconds. (";
    cout << framesDrawn/(watch.readMS()/1000.0) << " fps)\n";
    cout << obj->getVertexCount() << " vertexes, " ;
    cout << obj->getTriangleCount() << " triangles.\n";
    cout.flush();

    if (g_verbose>0) {
        cout << "Eye: " << ex << "," << ey << "," << ez;
        cout << " at angles " << lookleft << "," << lookup << endl;
    }

    return 0;
}
